/* SWITCHES.C - Switch-handling routines.
**
** Copyright (c)2007 by Ignacio R. Morelle "Shadow Master". (shadowm2006@gmail.com)
** Based on ModKeen 2.0.1 Copyright (c)2002-2004 Andrew Durdin. (andy@durdin.net)
** 
** This software is provided 'as-is', without any express or implied warranty.
** In no event will the authors be held liable for any damages arising from
** the use of this software.
** Permission is granted to anyone to use this software for any purpose, including
** commercial applications, and to alter it and redistribute it freely, subject
** to the following restrictions:
**    1. The origin of this software must not be misrepresented; you must not
**       claim that you wrote the original software. If you use this software in
**       a product, an acknowledgment in the product documentation would be
**       appreciated but is not required.
**    2. Altered source versions must be plainly marked as such, and must not be
**       misrepresented as being the original software.
**    3. This notice may not be removed or altered from any source distribution.
*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>

#include "utils.h"
#include "switches.h"

#include "pconio.h"

// GNU C compatibility:
#ifndef stricmp
#define stricmp strcasecmp
#endif /* !stricmp */

static SwitchStruct switches;

static void showswitches();
static void defaultswitches();
static int getswitch(char *string, char **option, char **value);

SwitchStruct *getswitches(int argc, char *argv[])
{
	char *option = NULL, *value = NULL;
	int i;
	
	defaultswitches();
	
	if(argc <= 1)
	{
		showswitches(argv[0]);
	}

	for(i = 1; i < argc; i++)
	{
		if(!getswitch(argv[i], &option, &value))
			quit("Invalid switch '%s'!", argv[i]);
		if(stricmp(option, "nowait") == 0)
		{
			NoWait = 1;
		}
		if(stricmp(option, "debug") == 0)
		{
			DebugMode = 1;
		}			
		else if(stricmp(option, "export") == 0)
		{
			if(switches.Import)
				quit("Cannot both import and export!");
			switches.Export = 1;
		}
		else if(stricmp(option, "import") == 0)
		{
			if(switches.Export)
				quit("Cannot both import and export!");
			switches.Import = 1;
		}
		else if(stricmp(option, "episode") == 0)
		{
			/* Possible values for --episode:
			 * 1-3: Keen 1-3 (episode 1-3 engine) Episode = 1-3
			 * 4-6: Keen 4-6 (episode 4-6 engine) Episode = 4-6
			 * D: Keen Dreams (episode 4-6 engine) Episode = 7
			 * DD2: Dangerous Dave 2 (episode 1-3 engine) Episode = 8
			 * Episode = 9 is reserved for future use
			 * BM1: Bio Menace 1 (episode 4-6 engine) Episode = 0xA
			 * BM2: Bio Menace 2 (episode 4-6 engine) Episode = 0xB
			 * BM3: Bio Menace 3 (episode 4-6 engine) Episode = 0xC
			 */
			if(!value)
				quit("No episode number given!");
			if(strcasecmp(value, "D") == 0) switches.Episode = 7;
			else if(strcasecmp(value, "DD2") == 0) switches.Episode = 8;
			else if(strcasecmp(value, "BM1") == 0) switches.Episode = 0xA;
			else if(strcasecmp(value, "BM2") == 0) switches.Episode = 0xB;
			else if(strcasecmp(value, "BM3") == 0) switches.Episode = 0xC; 
			else switches.Episode = atoi(value);
			
			// Remove when support for other Keen engine games is implemented
			if ((switches.Episode == 8)||((switches.Episode >= 0xA)&&(switches.Episode <= 0xC)))
				quit("Support for other Keen engine games has not been implemented yet!");			
			if(switches.Episode < 1 || switches.Episode > 7)
				quit("Episode must an integer between 1 and 6 or letter D (Keen Dreams)!");
			if (switches.Episode == 7)
				quit("Keen Dreams support not correctly implemented yet!");
		}
		else if(stricmp(option, "keendir") == 0)
		{
			if(!value)
				quit("No directory for Keen files given!");

			strncpy(switches.InputPath, value, PATH_MAX);
		}
		else if(stricmp(option, "bmpdir") == 0)
		{
			if(!value)
				quit("No directory for BMP files given!");

			strncpy(switches.OutputPath, value, PATH_MAX);
		}
		else if(stricmp(option, "backup") == 0)
		{
			switches.Backup = 1;
		}
		else if(stricmp(option, "help") == 0 || stricmp(option, "?") == 0)
		{
			showswitches();
			exit(0);
		}
		else
		{
			showswitches();
			if (strlen(option) > 0) quit("Unknown switch '%s'!", option);
			quit("No parameters given!");
		}
	}
	
	if(!switches.Import && !switches.Export)
		quit("Either -import or -export must be given!");
	if(switches.Episode == 0)
		quit("The episode number must be given!");
	
	return &switches;
}

static void defaultswitches()
{
	strncpy(switches.InputPath, ".", PATH_MAX);
	strncpy(switches.OutputPath, ".", PATH_MAX);
	
	switches.Backup = 0;
	switches.Episode = 0;
	switches.Export = 0;
	switches.Import = 0;
}

/* Switch format: -option="value string" -option -option=value */
static int getswitch(char *string, char **option, char **value)
{
	unsigned int len = strlen(string);
	char *p;
	
	/* An empty string is not a valid switch */
	if(len == 0)
		return 0;
		
	/* A switch must begin with '-' or '/' */
	p = string;
	if(*p != '-' && *p != '/')
		return 0;
		
	p++;
	/* For UNIX switches style compatibility */
	if (*p == '-') p++;
	if (!p) return 1;
	
	*option = p;
	*value = NULL;
	while(*p)
	{
		if(*p == '=' || *p == ':')
		{
			*p = 0;
			p++;
			
			if(*p != '"')
			{
				*value = p;
			}
			else
			{
				*value = p++;
				while(*p && *p != '"')
					p++;
				*p = 0;
			}
			
			/* We've found a valid switch */
			return 1;
		}
		p++;
	}
	/* We've found a valid switch -- but it has no value*/
	return 1;
}

static void showswitches(char* appname)
{
	fprintf(stderr, "Run \"%s -help\" for command line usage info.", appname);
}

